@@ -12,15 +12,20 @@ import enum
 import logging
 import re
 import time
+import traceback
 
-from typing import Optional, Callable, Coroutine, Any
+from typing import Optional, Callable, Coroutine, Any, Union
 
 from bleak import BleakScanner, BleakClient, BleakError
 from bleak import BleakError, BleakClient, BleakScanner
+from bleak.backends.bluezdbus.client import BleakClientBlueZDBus
 from bleak.backends.device import BLEDevice
 from bleak.backends.scanner import AdvertisementData
 
 import pyDE1.default_logger
+from pyDE1.bleak_client_wrapper import BleakClientWrapped
+from pyDE1.de1.exceptions import DE1APIValueError
+from pyDE1.dispatcher.resource import ConnectivityEnum
 from pyDE1.event_manager import EventPayload, SubscribedEvent
 from pyDE1.event_manager.events import ConnectivityState, ConnectivityChange
 from pyDE1.scale.events import ScaleWeightUpdate, ScaleButtonPress, \
@@ -46,8 +51,8 @@ class ScaleNotConnectedError(ScaleError):
 
 class Scale:
 
-    def __init__(self, address=None):
-        self._address: Optional[str] = address
+    def __init__(self):
+        self._address = None
         self._name: Optional[str] = None
 
         # These are often model-specific, override in subclass init
@@ -64,7 +69,7 @@ class Scale:
         self._tare_threshold = 0.05  # grams, within this, considered "at zero"
         self.hold_at_tare = False
 
-        self._bleak_client: Optional[BleakClient] = None
+        self._bleak_client: Optional[BleakClientWrapped] = None
 
         self._event_connectivity: SubscribedEvent = SubscribedEvent(self)
         self._event_weight_update: SubscribedEvent = SubscribedEvent(self)
@@ -108,13 +113,22 @@ class Scale:
 
     @property
     def address(self):
-        return self._address
+        addr = self._address
+        if isinstance(addr, BLEDevice):
+            addr = addr.address
+        return addr
 
-    @property
-    def address_is_persistent(self):
-        # macOS UUIDs don't seem to be persistent across Mac reboots
-        return bool(re.match(r'([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}$',
-                             self._address))
+    @address.setter
+    def address(self, address: Union[BLEDevice, str]):
+        if self.address is not None:
+            raise DE1APIValueError(
+                "Changing the scale address is not yet supported")
+        # If using BLEDevice directly
+        # AttributeError: 'BleakClientBlueZDBus' object has no attribute 'lower'
+        if isinstance(address, BleakClientBlueZDBus):
+            self._address = address.address
+        else:
+            self._address = address
 
     @property
     def name(self):
@@ -125,10 +139,12 @@ class Scale:
         return self._sensor_lag
 
     async def connect(self, timeout=5.0):
+        class_name = type(self).__name__
+        logger.info(f"Connecting to {class_name} at {self.address}")
         if self._bleak_client is None:
             if self.address is None:
                 raise ScaleNoAddressError
-            self._bleak_client = BleakClient(self.address)
+            self._bleak_client = BleakClientWrapped(self.address)
         self._bleak_client.set_disconnected_callback(
             self._create_disconnect_callback()
         )
@@ -138,14 +154,16 @@ class Scale:
             self._bleak_client.connect(timeout=timeout),
         )
         if self.is_connected:
-            self.register_atexit_disconnect()
-            logger.info(f"Connected to {type(self)} at {self.address}")
+            self._address = self._bleak_client.address
+            if self.name is None:
+                self._name = self._bleak_client.name
+            logger.info(f"Connected to {class_name} at {self.address}")
             await self._event_connectivity.publish(
                 ConnectivityChange(arrival_time=time.time(),
                                    state=ConnectivityState.CONNECTED))
         else:
             logger.error(
-                f"Connection failed to {type(self)} at {self.address}")
+                f"Connection failed to {class_name} at {self.address}")
             await self._event_connectivity.publish(
                 ConnectivityChange(arrival_time=time.time(),
                                    state=ConnectivityState.DISCONNECTED))
@@ -156,17 +174,23 @@ class Scale:
         if self.supports_button_press:
             await self.start_sending_button_updates()
         if not hold_notification:
-            await self._event_connectivity.publish(
-                ConnectivityChange(
-                    arrival_time=time.time(),
-                    state=ConnectivityState.READY
-                )
-            )
+            await self._notify_ready()
+
+    async def _notify_ready(self):
+        await self._event_connectivity.publish(
+            ConnectivityChange(arrival_time=time.time(),
+                               state=ConnectivityState.READY))
+
+    async def _notify_not_ready(self):
+        await self._event_connectivity.publish(
+            ConnectivityChange(arrival_time=time.time(),
+                               state=ConnectivityState.NOT_READY))
 
     async def disconnect(self):
-        logger.info(f"Disconnecting from {type(self)}")
+        class_name = type(self).__name__
+        logger.info(f"Disconnecting from {class_name}")
         if self._bleak_client is None:
-            logger.info(f"Disconnecting from {type(self)}; no client")
+            logger.info(f"Disconnecting from {class_name}; no client")
         await asyncio.gather(
             self._bleak_client.disconnect(),
             self._event_connectivity.publish(
@@ -175,18 +199,16 @@ class Scale:
         )
         if self.is_connected:
             logger.error(
-                f"Disconnect failed from {type(self)} at {self.address}")
+                f"Disconnect failed from {class_name} at {self.address}")
             await self._event_connectivity.publish(
                 ConnectivityChange(arrival_time=time.time(),
                                    state=ConnectivityState.CONNECTED))
         else:
-            logger.info(f"Disconnected from {type(self)} at {self.address}")
+            logger.info(f"Scale.disconnect(): Disconnected from {class_name} at {self.address}")
             await self._event_connectivity.publish(
                 ConnectivityChange(arrival_time=time.time(),
                                    state=ConnectivityState.DISCONNECTED))
 
-        # TODO: Unregister atexit disconnect
-
     # TODO: Decide how to handle  self._disconnected_callback
 
     @property
@@ -235,6 +257,7 @@ class Scale:
                 f"Tare request skipped, too soon, {dt:0.3f} seconds")
         return self._last_tare_request_sent
 
+
     async def _tare_internal(self):
         raise NotImplementedError
 
@@ -247,6 +270,18 @@ class Scale:
     async def display_off(self):
         raise NotImplementedError
 
+    # The two *_bool for API
+
+    async def tare_with_bool(self, do_it=True):
+        if do_it:
+            await self.tare()
+
+    async def display_bool(self, on):
+        if on:
+            await self.display_on()
+        else:
+            await self.display_off()
+
     @property
     def estimated_period(self):
         return self._estimated_period
@@ -326,9 +361,14 @@ class Scale:
     def _create_disconnect_callback(self) -> Callable:
         scale = self
 
-        def disconnect_callback(client: BleakClient):
+        def disconnect_callback(client: BleakClientWrapped):
             nonlocal scale
-            logger.info(f"Disconnected from {type(scale)} at {scale.address}")
+            class_name = type(self).__name__
+            logger.info(
+                "disconnect_callback: "
+                f"Disconnected from {class_name} at {client.address}, "
+                "willful_disconnect: "
+                f"{client.willful_disconnect}")
             asyncio.create_task(scale._event_connectivity.publish(
                 ConnectivityChange(arrival_time=time.time(),
                                    state=ConnectivityState.DISCONNECTED)))
@@ -393,21 +433,26 @@ class Scale:
                     self._n_counter = 0
                     logger.debug(f"Scale period: {self._ma}")
 
-    def atexit_disconnect(self):
 
-        def sync_disconnect():
-            nonlocal self
-            if not self.is_connected:
-                logger.debug(
-                    f"atexit sync_disconnect: Not connected to {self}")
-                return
-            else:
-                logger.info(
-                    f"atexit sync_disconnect: Disconnecting {self}")
-                loop = asyncio.get_event_loop()
-                loop.run_until_complete(self.disconnect())
+    # TODO: Deal with connectivity as a mixin
 
-        return sync_disconnect
+    # For API
+    @property
+    def connectivity(self):
+        if self.is_connected:
+            retval = ConnectivityEnum.CONNECTED
+        else:
+            # intentionally vague, as PUT "connecting" isn't needed
+            retval = ConnectivityEnum.NOT_CONNECTED
+        return retval
 
-    def register_atexit_disconnect(self):
-        atexit.register(self.atexit_disconnect())
+    async def connectivity_setter(self, value: ConnectivityEnum):
+        assert isinstance(value, ConnectivityEnum), \
+            f"mode of {value} not a ConnectivityEnum " \
+            "in Scale.connectivity_setter()"
+        if value is ConnectivityEnum.CONNECTED \
+                and not self.is_connected:
+            await self.connect()
+        elif value is ConnectivityEnum.NOT_CONNECTED \
+                and self.is_connected:
+            await self.disconnect()
\ No newline at end of file
